import argparse
import inspect
import logging
import os
import sys
from pathlib import Path

support_types = ('str', 'int', 'bool', 'float')

from cpbox.tool import datatypes
from cpbox.tool import system

class BaseCliApp(object):

    def __init__(self):
        self.cli_root_dir = self._get_cli_root_dir()
        from tibase.core.rootcontext import gcontext
        self.logger = logging.getLogger(self.__class__.__name__)

    def _get_cli_root_dir(self):
        if hasattr(sys.modules['__main__'], '__file__'):
            return Path(sys.argv[0]).resolve().parent
        else:
            return Path.cwd()

    def run_cmd_ret(self, cmd, log=True):
        return self.run_cmd(cmd, log=log)[1]

    def run_cmd(self, cmd, log=True):
        if log:
            self.logger.info('run_cmd: %s', cmd)
        return system.run_cmd(cmd)

    def shell_run(self, cmd, keep_pipeline=True, exit_on_error=True, dry_run=False, log=True):
        if log:
            self.logger.info('shell_run: %s', cmd)
        if dry_run:
            return 0
        return system.shell_run(cmd, keep_pipeline=keep_pipeline, exit_on_error=exit_on_error)

def append_cli(container, cli):
    cli_map = {}
    if hasattr(container, '__cli_map__'):
        cli_map = container.__cli_map__
    for key, attr in cli.__class__.__dict__.items():
        if key[0:1] != '_' and callable(attr):
            cli_map[key] = cli
            setattr(container, key, attr.__get__(cli))

def append_cli_for_composition(container, cli):
    return append_cli(container, cli)

def merge_class(*proxy_cls_list):
    def real_decorator(cls):
        cls.__proxy_src_cls__ = cls
        cls.__proxy_cls_list__ = list(proxy_cls_list)
        return cls
    return real_decorator

def keep_class(cls):
    cls.keep_class_for_cli = True
    return cls

def keep_method(f):
    f.keep_method_for_cli = True
    return f

def keep_class_for_inheritance(cls):
    return keep_class(cls)

def keep_method_for_inheritance(f):
    return keep_method(f)

def merge_class_for_composition(*proxy_cls_list):
    return merge_class(*proxy_cls_list)

class AppSepc(object):

    def __init__(self, app, base_args_option=None):
        if base_args_option is None:
            base_args_option = _build_base_args_option_from_init(app)
        self.base_args_option = base_args_option

        self.app = app
        self.get_app_public_methods(app)
        self.method_spec_map = self._get_app_method_spec()
        self.num_to_name_map = {str(index): name for index, name in enumerate(self.public_methods)}

        self.method_will_call = None

    def get_app_public_methods_for_proxy_class(self, app):
        proxy_src_cls = app.__dict__['__proxy_src_cls__']
        proxy_cls_list = app.__dict__['__proxy_cls_list__']

        method_names = []
        method_map = {}

        src_cls_methods = get_cls_cli_methods(proxy_src_cls, True)
        method_names.extend(src_cls_methods.keys())
        method_map.update(src_cls_methods)

        for cls in proxy_cls_list:
            cls_methods = get_cls_cli_methods(cls, False)
            method_names.extend(cls_methods.keys())
            method_map.update(cls_methods)

        self.public_methods = method_names
        self.method_map = method_map

    def get_app_public_methods_for_simple_class(self, app):
        cls_methods = get_cls_cli_methods(app, True)
        self.public_methods = cls_methods.keys()
        self.method_map = cls_methods

    def get_app_public_methods(self, app):
        if '__proxy_cls_list__' in app.__dict__:
            self.get_app_public_methods_for_proxy_class(app)
        else:
            self.get_app_public_methods_for_simple_class(app)

    def find_method_name(self, num_or_name):
        if num_or_name in self.num_to_name_map:
            return self.num_to_name_map[num_or_name]
        if num_or_name in self.method_spec_map:
            return num_or_name
        return None

    def fix_terminal_columns(self):
        if os.name == 'nt':
            return
        if sys.stdin.isatty() and sys.stdout.isatty():
            rows, columns = os.popen('stty size', 'r').read().split()
            os.environ['COLUMNS'] = columns

    def resolve(self, default_method=None):
        self.fix_terminal_columns()

        method_will_call = None
        if len(sys.argv) > 1:
            method_will_call = sys.argv[1].lstrip('-')
            method_will_call = method_will_call.replace('-', '_')
        if method_will_call is None and default_method is not None:
            method_will_call =  default_method
        method_will_call = self.find_method_name(method_will_call)

        if not method_will_call:
            self.show_help_then_exit()

        self.method_will_call = method_will_call
        self._parse_args()

    def _parse_args(self):
        method_will_call = self.method_will_call
        method_spec = self.method_spec_map[method_will_call]
        kwargs = build_argparser_for_func(method_will_call, method_spec)

        self.base_kwargs = {}
        if self.base_args_option:
            common_args_keys = set(self.base_args_option.get('args'))
            methods_args_keys = set(method_spec['method_origin_args'])
            method_kwargs = {}
            for key, value in kwargs.items():
                if key in common_args_keys:
                    self.base_kwargs[key] = value
                if key in methods_args_keys:
                    method_kwargs[key] = value
            self.kwargs = method_kwargs
        else:
            self.kwargs = kwargs

    def run(self):
        kwargs = self.kwargs
        _app = self.app(**self.base_kwargs)
        func = getattr(_app, self.method_will_call)
        func(**kwargs)

    def _get_app_method_spec(self):
        method_spec_map = {}
        for method_name in self.public_methods:
            method_spec_map[method_name] = self.get_method_spec(method_name)
        return method_spec_map

    def get_method_spec(self, method_name):
        method_attr = self.method_map[method_name]
        method_arg_spec = inspect.getfullargspec(method_attr)
        extracted = _extract_args_spec(method_arg_spec)
        default_values = extracted.get('default_values')
        args = extracted.get('args', [])

        method_spec = {}
        method_spec['method_origin_args'] = list(args)
        if self.base_args_option != None:
            args.extend(self.base_args_option.get('args', []))
            args = list(set(args))
            default_values_from_common_args = self.base_args_option.get('default_values', [])
            if default_values is None:
                default_values = default_values_from_common_args
            else:
                default_values.update(default_values_from_common_args)

        method_spec['args'] = args
        method_spec['default_values'] = default_values
        method_spec['parameters'] = _check_parameters_spec(method_name, args, default_values)
        return method_spec

    def show_help_then_exit(self):
        parser = argparse.ArgumentParser(add_help=False)
        parser.allow_abbrev = False
        parser.add_argument('-h', '--help', action='store_true', help='show this help message and exit')
        for index, method_name in enumerate(self.public_methods):
            method_spec = self.method_spec_map.get(method_name)
            help_text = self._build_help_for_method(method_name, method_spec)
            parser.add_argument('--' + method_name, action='store_true', help=help_text)
        parser.print_help()
        if any(arg in ('-h', '--help') for arg in sys.argv[1:]):
            sys.exit(0)
        sys.exit(1)

    def _build_help_for_method(self, method_name, method_spec):
        if not method_spec:
            return ''
        parameters = method_spec.get('parameters', {})
        default_values = method_spec.get('default_values') or {}
        args = method_spec.get('args', [])
        parts = []
        for arg in args:
            param = parameters.get(arg, {})
            arg_type = param.get('type', 'str')
            if arg in default_values:
                parts.append(f"{arg}[{arg_type}={default_values.get(arg)}]")
            else:
                parts.append(f"{arg}[{arg_type}*]")
        if not parts:
            return ''
        return ', '.join(parts)

def _is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def _check_parameter_item_default_value(arg_name, parameter_item, default_values):
    if default_values is None or arg_name not in default_values:
        return parameter_item

    value = default_values[arg_name]
    type = _determine_type(value)

    if type is None:
        raise Error('Can not determine type for parameter: %s, parameter_item: %s' %s (arg_name, parameter_item))

    parameter_item['type'] = type
    parameter_item['des'] = 'type: %s, default value is: %s' % (type, value)
    return parameter_item

def _determine_type(value):
    type = None
    if value is None:
        type = 'str'
    elif isinstance(value, bool):
        type = 'bool'
    elif isinstance(value, int):
        type = 'int'
    elif isinstance(value, str):
        type = 'str'
    elif _is_float(value):
        type = 'float'
    return type

def _convert_value_for_type(value, type):
    if value is None:
        return value
    elif type == 'bool':
        if isinstance(value, str):
            try:
                return _strtobool(value)
            except:
                return False
        return bool(value)
    elif type == 'int':
        return int(value)
    elif type == 'str':
        return str(value)
    elif type == 'float':
        return float(value)
    return value

def _check_parameters_spec(method_name, args, default_values):
    parameters = {}
    for arg_name in args:
        parameter_item = {}
        if isinstance(parameter_item, str):
            parameter_item = {
                    'type': 'str',
                    'des': parameter_item
                    }
        parameter_item = _check_parameter_item_default_value(arg_name, parameter_item, default_values)
        if 'type' not in parameter_item:
            parameter_item['type'] = 'str'
        if 'des' not in parameter_item:
            parameter_item['des'] = arg_name

        parameters[arg_name] = parameter_item
        _check_parameter_item(arg_name, parameter_item)

    return parameters

def _check_parameter_item(arg_name, parameter_item):
    if 'type' not in parameter_item:
        print('Missing type in parameter item: %s' % parameter_item)
    if parameter_item['type'] not in support_types:
        print('The type: %s is not in support types: %s' % (parameter_item['type'], support_types))
    if not parameter_item['des']:
        print('Missing parameter descripion(`des`): %s' % parameter_item)

def get_cls_cli_methods(cls, exclude_parent=True):
    methods = get_cls_public_methods(cls)
    if exclude_parent:
        parenet_cls_methods = get_cls_parent_methods_for_exclude(cls)
        return {key: value for key, value in methods.items() if key not in parenet_cls_methods}
    else:
        return methods

def get_cls_public_methods(cls):
    list = inspect.getmembers(cls, predicate=lambda x: inspect.isfunction(x) or inspect.ismethod(x))
    methods = {item[0]: item[1] for item in list if item[0][0:1] != '_'}
    return methods

def get_cls_parent_methods_for_exclude(cls):
    methods = {}
    for parent_cls in cls.__bases__:
        if parent_cls.__dict__.get('keep_class_for_cli', False):
            continue
        parent_methods = get_cls_public_methods(parent_cls)
        parent_methods = {key: val for key, val in parent_methods.items() if not val.__dict__.get('keep_method_for_cli', False)}
        methods.update(parent_methods)
    return methods

def build_argparser_for_func(method_name, method_spec):

    args = method_spec['args']
    default_values = method_spec['default_values']

    parser = argparse.ArgumentParser()
    parser.add_argument('--' + method_name, action='store_true')

    parameters = method_spec.get('parameters', {})
    for arg_name in args:
        parameter = parameters.get(arg_name)
        help = parameter.get('des', '')
        metavar = ''

        if default_values is not None and arg_name in default_values:
            if len(arg_name) == 1:
                parser.add_argument('-' + arg_name, metavar=metavar, help=help, default=default_values[arg_name], nargs=1, required=False)
            else:
                parser.add_argument('--' + arg_name, metavar=metavar, help=help, default=default_values[arg_name], nargs=1, required=False)
        else:
            if len(arg_name) == 1:
                parser.add_argument('-' + arg_name, metavar=metavar, help=help, nargs=1, required=True)
            else:
                parser.add_argument('--' + arg_name, metavar=metavar, help=help, nargs=1, required=True)
    args_input, _ = parser.parse_known_args()
    kwargs = {}
    for arg_name in args:
        parameter = parameters.get(arg_name)
        attr = getattr(args_input, arg_name)
        if default_values is not None and arg_name in default_values and attr == default_values[arg_name]:
            kwargs[arg_name] = attr
        else:
            kwargs[arg_name] = _convert_value_for_type(
                    _get_first_from_list_or_item_self(attr), parameter['type'])
    return kwargs

def _get_first_from_list_or_item_self(attr):
    try:
        return next(iter(attr), None)
    except TypeError:
        return attr

def _strtobool(value):
    value = value.strip().lower()
    if value in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    if value in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    raise ValueError('invalid truth value: %s' % value)

def _check_args_has(args, option):
    ret = getattr(args, option)
    return ret

def _extract_args_spec(arg_spec):
    args = [arg for arg in arg_spec.args if arg != 'self']
    default_values = {}
    required_args = []
    if arg_spec.defaults:
        default_values = dict(zip(arg_spec.args[-len(arg_spec.defaults):], arg_spec.defaults))
    for arg in args:
        if arg not in default_values:
            required_args.append(arg)
    return {
        'args': args,
        'default_values': default_values,
        'required_args': required_args,
    }

def _build_base_args_option_from_init(app_cls):
    init_spec = inspect.getfullargspec(app_cls.__init__)
    return _extract_args_spec(init_spec)

def run_app(app, default_method=None, base_args_option=None, **kwargs):
    app_spec = AppSepc(app, base_args_option)
    app_spec.resolve(default_method)
    app_spec.run()



from typing import Dict, Type, Any

class Singleton:
    _instances: Dict[Type, Any] = {}
    
    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__new__(cls)
        return cls._instances[cls]
    
    def __init__(self):
        # 确保初始化代码只执行一次
        if not hasattr(self, '_initialized'):
            self._initialized = True
            self._init_once()
    
    def _init_once(self):
        """子类可以重写此方法来进行一次性初始化"""
        pass


import os
import sys
from cpbox.tool import file
from pathlib import Path
from tibase.core import tibaselog

from dotenv import dotenv_values

import logging
logger = logging.getLogger(__name__)

def detect_base_dir(base_dir=Path.cwd()):
    # 先从当前目录开始，如果当前目录有 .env 文件，则返回当前目录
    # 否则，向上遍历目录，直到找到 .env 文件或到达根目录
    # 如果运行脚本，则从脚本所在目录开始
    base_dir = Path.cwd()
    if hasattr(sys.modules['__main__'], '__file__'):
        base_dir = Path(sys.argv[0]).resolve().parent
    while True:
        if (base_dir / '.env').exists():
            return base_dir
        elif base_dir == base_dir.parent:
            return None
        else:
            base_dir = base_dir.parent

def auto_load_dot_env(base_dir: Path):
    env_dict = dotenv_values(base_dir / '.env')
    if (base_dir / '.env.local').exists():
        env_local_dict = dotenv_values(base_dir / '.env.local')
        env_dict.update(env_local_dict)

    for key, value in env_dict.items():
        if value is None:
            continue
        if key in os.environ:
            continue
        os.environ[key] = value

    return env_dict


class EnvConfig(dict):

    def __init__(self, data):
        dict.__init__(self, data or {})

    def _to_type(self, raw, item_type, default, strict):
        if raw is None:
            return default
        if item_type is bool:
            value = str(raw).lower()
            if value in ('true', '1', 'yes', 'y', 'on'):
                return True
            if value in ('false', '0', 'no', 'n', 'off'):
                return False
            return default
        try:
            return item_type(raw)
        except ValueError:
            if strict:
                raise
            return default

    def get_bool(self, key, default=False):
        return self._to_type(self.get(key, None), bool, default, strict=False)

    def get_int(self, key, default=0, strict=True):
        return self._to_type(self.get(key, None), int, default, strict)

    def get_float(self, key, default=0.0, strict=True):
        return self._to_type(self.get(key, None), float, default, strict)

    def get_list(self, key, item_type=str, sep=',', default=None):
        raw = self.get(key, None)
        if raw is None:
            return default if default is not None else []
        items = str(raw).split(sep)
        items = [item.strip() for item in items]
        items = [item for item in items if item]
        return [self._to_type(item, item_type, default, strict=True) for item in items]


def _detect_app_env(env_config):
    app_env = os.getenv('APP_ENV') or env_config.get('APP_ENV') or env_config.get('ENV') or 'prod'
    app_env = str(app_env).lower()
    if app_env not in ('dev', 'test', 'pre', 'prod'):
        return 'prod'
    return app_env


class _GlobalInfo:

    def __init__(self):
        self._root_dir = detect_base_dir()
        self._env_config = EnvConfig(auto_load_dot_env(self._root_dir))

_global_info = _GlobalInfo()

from tibase.core import cls
class GContext(cls.Singleton):

    def __init__(self):
        super().__init__()

    def _init_once(self):
        self.root_dir = _global_info._root_dir
        self.env_config = _global_info._env_config
        self.app_env = _detect_app_env(self.env_config)
        self.env_config['APP_ENV'] = self.app_env
        self.is_dev_mode = self.app_env in ('dev', 'test')

        self._init_logger()

        self.output_dir = self.root_dir / 'output'
        file.ensure_dir(self.output_dir)

        self.workspace_dir = self.output_dir / 'workspace'
        file.ensure_dir(self.workspace_dir)

    def _init_logger(self):
        tibaselog.log_manager.init_from_env(self.env_config)
        logs_dir = self.root_dir / 'output' / 'logs'
        app_name = self.env_config.get('APP_NAME', 'tiapp')
        tibaselog.log_manager.add_file_logger(app_name, logs_dir)

    # ==================== 延迟加载的工具模块 ====================

    @property
    def tools(self):
        """延迟加载工具模块"""
        from tibase.core import tools
        return tools

    @property
    def log_mgr(self):
        return tibaselog.log_manager

gcontext = GContext()



"""
常用工具函数
提供 YAML、JSON 等常用操作的便捷方法
"""
import os
from io import StringIO
from cpbox.tool import file
from tibase.core import cls


class YAMLInitializer(cls.Singleton):
    """YAML 实例初始化器（使用 SIP 模式）"""

    def _init_once(self):
        """初始化 YAML 实例，只执行一次"""
        from ruamel.yaml import YAML
        self.yaml = YAML(typ='rt')
        self.yaml.indent(mapping=2, sequence=4, offset=2)
        self.yaml.allow_unicode = True
        self.yaml.sort_keys = False
        self.yaml.default_flow_style = False
        self.yaml.default_style = None

yaml = YAMLInitializer().yaml

def pyaml(data) -> str:
    """将数据转换为 YAML 字符串"""
    stream = StringIO()
    yaml.dump(data, stream=stream)
    return stream.getvalue()


def pjson(data):
    """将数据转换为 JSON 字符串（延迟加载）"""
    import json
    return json.dumps(data, ensure_ascii=False, indent=2)


def load_yaml_file(fn, fallback=None):
    """从 YAML 文件加载数据（延迟加载）"""
    if not os.path.isfile(fn):
        return fallback
    data = None
    with open(fn, 'r', encoding='utf-8') as f:
        data = yaml.load(f)
    return data if data is not None else fallback


def dump_yaml_file(fn, data):
    """将数据写入 YAML 文件（延迟加载）"""
    file.ensure_dir_for_file(str(fn))
    with open(fn, 'w', encoding='utf-8') as outfile:
        yaml.dump(data, outfile)

def load_json_file(fn, fallback=None):
    """从 JSON 文件加载数据（延迟加载）"""
    import json
    if not os.path.isfile(fn):
        return fallback
    try:
        with open(fn, 'r', encoding='utf-8') as f:
            data = json.load(f)
        return data if data is not None else fallback
    except Exception:
        return fallback


def dump_json_file(fn, data):
    """将数据写入 JSON 文件（延迟加载）"""
    import json
    file.ensure_dir_for_file(str(fn))
    with open(fn, 'w', encoding='utf-8') as f:
        json.dump(data, f, ensure_ascii=False, indent=2)


import time
from datetime import datetime
from typing import Dict, Type, Any

class IDMan:

    def __init__(self):
        self.last_time = 0
        self.sequence = 0
    
    def next_id(self) -> int:
        now = int(time.time() * 1000)  # 当前毫秒时间戳
        
        # 如果时间戳相同，增加序列号
        if now == self.last_time:
            self.sequence += 1
        else:
            # 时间戳不同，重置序列号
            self.last_time = now
            self.sequence = 0
        
        # 组合ID：时间戳左移4位 + 序列号
        id = (now << 4) | self.sequence
        return id

def explain_id(id_value: int) -> str:
    # 提取时间戳和序列号
    timestamp = id_value >> 4
    sequence = id_value & 0xF  # 获取低4位
    
    # 转换时间戳为可读格式
    dt = datetime.fromtimestamp(timestamp / 1000)  # 转换为秒
    time_str = dt.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]  # 保留毫秒
    
    return f"ID: {id_value} (时间: {time_str}, 序列号: {sequence})"

class IDTag:

    def __init__(self, get_tag_func):
        self.get_tag_func = get_tag_func

    def __str__(self):
        return f"{self.get_tag_func()}"

class IDClass:
    id_man = IDMan()  # 类成员，只初始化一次
    
    def __init__(self, get_tag_func=None):
        self.id = self.id_man.next_id()
        self.base_tag = f"{self.__class__.__name__}/{self.id}"
        if hasattr(self, "_get_tag"):
            self.tag = IDTag(self._get_tag)
        else:
            self.tag = self.base_tag
    
    def explain_id(self):
        return explain_id(self.id)